use crate::lib::entity::{Ant, Egg, Queen};
use crate::lib::point::{astar, Point};
use crate::lib::screen::{BoardCommand, Screen};
use crate::lib::world::World;
use rand::prelude::SliceRandom;
use rand::thread_rng;

pub trait AI {
    fn step(&mut self, b: &Screen, w: &mut World) -> BoardCommand;
    fn get_position(&self) -> Point;
    fn plan(&mut self, w: &World) {}
}

pub enum AIGoal {
    Reach(Point),
    Idle,
}

impl AI for Ant {
    fn get_position(&self) -> Point {
        self.pos.clone()
    }

    fn plan(&mut self, w: &World) {
        if self.plan.len() == 0 {
            self.plan = match self.goal {
                AIGoal::Reach(target) => {
                    if &self.pos == &target {
                        self.goal = AIGoal::Idle;
                        vec![]
                    } else {
                        astar(&self.pos, &target)
                    }
                },
                AIGoal::Idle => vec![],
            }
        }
    }

    // return the next move for this ant
    fn step(&mut self, b: &Screen, w: &mut World) -> BoardCommand {
        let choice = match self.goal {
            AIGoal::Idle => {
               let valid = w.get_valid_movements(&self.pos, b);
               let mut rng = thread_rng();
               valid.choose(&mut rng).cloned()      
            },
            AIGoal::Reach(_) => {
                let movement = self.plan.pop();
                movement
            }
        };
        

        if !choice.is_none() {
            let pos = choice.unwrap();
            if w.cleared.contains(&pos) {
                let old_pos = self.pos;
                self.pos = pos;
                w.update_occupied(&old_pos, &pos);
            } else {
                w.clear(pos);
                match self.goal {
                    // push plan back because we haven't actually moved towards it
                    AIGoal::Reach(_) => {
                        self.plan.push(pos);
                    }
                    AIGoal::Idle => {}
                }
            }
        }

        BoardCommand::Noop
    }
}

impl AI for Egg {
    fn step(&mut self, b: &Screen, w: &mut World) -> BoardCommand {
        if self.counter > 0 {
            self.counter -= 1;
            return BoardCommand::Noop;
        } else {
            BoardCommand::Hatch(self.id, self.queen_id)
        }
    }

    fn get_position(&self) -> Point {
        self.pos.clone()
    }
}

impl AI for Queen {
    fn get_position(&self) -> Point {
        self.pos.clone()
    }
    fn step(&mut self, b: &Screen, w: &mut World) -> BoardCommand {
        let valid: Vec<Point> = w.get_valid_movements(&self.pos, b);
        let mut rng = thread_rng();

        let choice = valid.choose(&mut rng).cloned();

        if !choice.is_none() {
            let pos = choice.unwrap();
            if w.cleared.contains(&pos) {
                // choose between laying an egg and moving
                if self.egg_count < 3 {
                    self.egg_count += 1;
                    return BoardCommand::LayEgg(pos, self.id);
                } else {
                    let old_pos = self.pos;
                    self.pos = pos;
                    w.update_occupied(&old_pos, &pos);
                }
            }
        }
        BoardCommand::Noop
    }
}
